@@ -210,7 +210,18 @@ int mnt_table_remove_fs(struct libmnt_table *tb, struct libmnt_fs *fs)
  * @tb: mountinfo file (/proc/self/mountinfo)
  * @root: returns pointer to the root filesystem (/)
  *
- * Returns: 0 on success or -1 case of error.
+ * The function uses parent ID from mountinfo file to determine root filesystem
+ * (the filesystem with the smallest ID). The function is designed mostly for
+ * applications where is necessary to sort mountpoints by IDs to get the tree
+ * of the mountpoints (e.g. findmnt default output).
+ *
+ * If you're not sure than use
+ *
+ *	mnt_table_find_target(tb, "/", MNT_ITER_BACKWARD);
+ *
+ * this is more robust and usable for arbitrary tab file (including fstab).
+ *
+ * Returns: 0 on success or less then zero case of error.
  */
 int mnt_table_get_root_fs(struct libmnt_table *tb, struct libmnt_fs **root)
 {
@@ -456,13 +467,18 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
 			return fs;
 	}
 
-	/* non-canonicaled path in struct libmnt_table */
+	/* non-canonicaled path in struct libmnt_table
+	 * -- note that mountpoint in /proc/self/mountinfo is already
+	 *    canonicalized by kernel
+	 */
 	mnt_reset_iter(&itr, direction);
 	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
 		char *p;
 
-		if (!fs->target || mnt_fs_is_swaparea(fs) ||
-		    (*fs->target == '/' && *(fs->target + 1) == '\0'))
+		if (!fs->target
+		    || mnt_fs_is_swaparea(fs)
+		    || mnt_fs_is_kernel(fs)
+		    || (*fs->target == '/' && *(fs->target + 1) == '\0'))
 		       continue;
 
 		p = mnt_resolve_path(fs->target, tb->cache);
@@ -703,6 +719,40 @@ struct libmnt_fs *mnt_table_find_pair(struct libmnt_table *tb, const char *sourc
 	return NULL;
 }
 
+/**
+ * mnt_table_find_devno
+ * @tb: /proc/self/mountinfo
+ * @devno: device number
+ * @direction: MNT_ITER_{FORWARD,BACKWARD}
+ *
+ * Note that zero could be valid device number for root pseudo filesystem (e.g.
+ * tmpfs).
+ *
+ * Returns: a tab entry or NULL.
+ */
+struct libmnt_fs *mnt_table_find_devno(struct libmnt_table *tb,
+				       dev_t devno, int direction)
+{
+	struct libmnt_fs *fs = NULL;
+	struct libmnt_iter itr;
+
+	assert(tb);
+
+	if (!tb)
+		return NULL;
+
+	DBG(TAB, mnt_debug_h(tb, "lookup DEVNO: %d", (int) devno));
+
+	mnt_reset_iter(&itr, direction);
+
+	while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
+		if (mnt_fs_get_devno(fs) == devno)
+			return fs;
+	}
+
+	return NULL;
+}
+
 /*
  * tb: /proc/self/mountinfo
  * fs: filesystem
@@ -837,8 +887,18 @@ static int is_mountinfo(struct libmnt_table *tb)
  * @tb: /proc/self/mountinfo file
  * @fstab_fs: /etc/fstab entry
  *
- * Checks if the @fstab_fs entry is already in the @tb table. The "swap"
- * is ignored.
+ * Checks if the @fstab_fs entry is already in the @tb table. The "swap" is
+ * ignored. This function explicitly compares source, target and root of the
+ * filesystems.
+ *
+ * Note that source and target are canonicalized only if a cache for @tb is
+ * defined (see mnt_table_set_cache()). The target canonicalization may
+ * triggers automount on autofs mountpoints!
+ *
+ * Don't use it if you want to know if a device is mounted, just use
+ * mnt_table_find_source() for the device.
+ *
+ * This function is designed mostly for "mount -a".
  *
  * TODO: check for loopdev (see mount/mount.c is_fstab_entry_mounted().
  *
@@ -847,14 +907,14 @@ static int is_mountinfo(struct libmnt_table *tb)
 int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 {
 	char *root = NULL;
-	const char *src = NULL;
-	char *xsrc = NULL, *tgt;
+	const char *src = NULL, *tgt = NULL;
+	char *xtgt = NULL;
 	int rc = 0;
 
 	assert(tb);
 	assert(fstab_fs);
 
-	if (mnt_fs_is_swaparea(fstab_fs))
+	if (mnt_fs_is_swaparea(fstab_fs) || mnt_table_get_nents(tb) == 0)
 		return 0;
 
 	if (is_mountinfo(tb)) {
@@ -870,15 +930,13 @@ int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 			src = mnt_fs_get_srcpath(fs);
 	}
 
-	if (src)
-		src = xsrc = mnt_resolve_spec(src, tb->cache);
-	else if (mnt_fs_is_pseudofs(fstab_fs))
+	if (!src)
 		src = mnt_fs_get_source(fstab_fs);
-	else
-		src = xsrc = mnt_resolve_spec(mnt_fs_get_source(fstab_fs),
-					      tb->cache);
 
-	tgt = mnt_resolve_path(mnt_fs_get_target(fstab_fs), tb->cache);
+	if (src && tb->cache && !mnt_fs_is_pseudofs(fstab_fs))
+		src = mnt_resolve_spec(src, tb->cache);
+
+	tgt = mnt_fs_get_target(fstab_fs);
 
 	if (tgt && src) {
 		struct libmnt_iter itr;
@@ -886,31 +944,36 @@ int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 
 		mnt_reset_iter(&itr, MNT_ITER_FORWARD);
 
-		while(mnt_table_next_fs(tb, &itr, &fs) == 0) {
+		while (mnt_table_next_fs(tb, &itr, &fs) == 0) {
+
+			if (!mnt_fs_streq_srcpath(fs, src))
+				continue;
 
 			if (root) {
-				/* mountinfo: compare root, source and target */
 				const char *r = mnt_fs_get_root(fs);
-
-				if (r && strcmp(r, root) == 0 &&
-				    mnt_fs_streq_srcpath(fs, src) &&
-				    mnt_fs_streq_target(fs, tgt))
-					break;
+				if (!r || strcmp(r, root) != 0)
+					continue;
 			}
-			/* mtab: compare source and target */
-			else if (mnt_fs_streq_srcpath(fs, src) &&
-				 mnt_fs_streq_target(fs, tgt))
+
+			/*
+			 * Compare target, try to minimize number of situations
+			 * when we need to canonicalize the path to avoid
+			 * readlink() on mountpoints.
+			 */
+			if (!xtgt) {
+				if (mnt_fs_streq_target(fs, tgt))
+					break;
+				if (tb->cache)
+					xtgt = mnt_resolve_path(tgt, tb->cache);
+			}
+			if (xtgt && mnt_fs_streq_target(fs, xtgt))
 				break;
+
 		}
 		if (fs)
 			rc = 1;		/* success */
 	}
 
-	if (xsrc && !tb->cache)
-		free(xsrc);
-	if (!tb->cache)
-		free(tgt);
-
 	free(root);
 	return rc;
 }
@@ -1055,11 +1118,16 @@ int test_find_pair(struct libmnt_test *ts, int argc, char *argv[])
 {
 	struct libmnt_table *tb;
 	struct libmnt_fs *fs;
+	struct libmnt_cache *mpc = NULL;
 	int rc = -1;
 
 	tb = create_table(argv[1]);
 	if (!tb)
 		return -1;
+	mpc = mnt_new_cache();
+	if (!mpc)
+		goto done;
+	mnt_table_set_cache(tb, mpc);
 
 	fs = mnt_table_find_pair(tb, argv[2], argv[3], MNT_ITER_FORWARD);
 	if (!fs)
@@ -1069,6 +1137,7 @@ int test_find_pair(struct libmnt_test *ts, int argc, char *argv[])
 	rc = 0;
 done:
 	mnt_free_table(tb);
+	mnt_free_cache(mpc);
 	return rc;
 }
 
@@ -1077,6 +1146,7 @@ static int test_is_mounted(struct libmnt_test *ts, int argc, char *argv[])
 	struct libmnt_table *tb = NULL, *fstab = NULL;
 	struct libmnt_fs *fs;
 	struct libmnt_iter *itr = NULL;
+	struct libmnt_cache *mpc = NULL;
 	int rc;
 
 	tb = mnt_new_table_from_file("/proc/self/mountinfo");
@@ -1093,6 +1163,11 @@ static int test_is_mounted(struct libmnt_test *ts, int argc, char *argv[])
 	if (!itr)
 		goto done;
 
+	mpc = mnt_new_cache();
+	if (!mpc)
+		goto done;
+	mnt_table_set_cache(tb, mpc);
+
 	while(mnt_table_next_fs(fstab, itr, &fs) == 0) {
 		if (mnt_table_is_fs_mounted(tb, fs))
 			printf("%s already mounted on %s\n",
@@ -1108,6 +1183,7 @@ static int test_is_mounted(struct libmnt_test *ts, int argc, char *argv[])
 done:
 	mnt_free_table(tb);
 	mnt_free_table(fstab);
+	mnt_free_cache(mpc);
 	mnt_free_iter(itr);
 	return rc;
 }
